import io
import pathlib
import sys
from typing import Literal

import tyro

from .core import Memos, normalize_logseq_content, parse_memos_from_markdown, render_site
from .templates import create_template_env


def build(
    source: pathlib.Path | Literal["-"],
    /,
    output: pathlib.Path | Literal["-"] = "-",
) -> None:
    """
    Build HTML from memos.

    Args:
        source: Path to source file or directory containing memos,
          or '-' for standard input (stdin)
        output: Output file path, or '-' for standard output (stdout)

    """
    # Load memos data
    match source:
        case "-":
            memos = parse_memos_from_markdown(sys.stdin.read())
        case path if path.suffix.lower() == ".md":
            memos = parse_memos_from_markdown(source.read_text())
        case path if path.is_dir():
            memos = parse_memos_from_directory(path)
        case _:
            msg = f"Unrecognized source (must be stdin, a .md file, or a directory: {source}"
            raise ValueError(msg)

    # Create template environment
    project_root = pathlib.Path.cwd()
    template_env = create_template_env(project_root)

    # Render site
    html = render_site(memos, template_env)

    # Output
    if output == "-":
        sys.stdout.write(html)
    else:
        output.write_text(html)


def main() -> None:
    tyro.cli(build)


def parse_memos_from_directory(directory: pathlib.Path) -> Memos:
    """Assemble a single stream from all files in a directory."""
    single_stream = io.StringIO()

    frontmatter_file = directory / "site.yaml"
    if not frontmatter_file.exists():
        msg = f"No site configuration found in {directory}, create and populate {directory}/site.yaml."
        raise ValueError(msg)
    single_stream.write("---\n")
    single_stream.write(frontmatter_file.read_text())
    single_stream.write("\n---\n")

    normalized_files = (
        normalize_logseq_content(journal.stem, journal.read_text())
        for journal in directory.glob("journals/[0-9][0-9][0-9][0-9][-_][0-9][0-9][-_][0-9][0-9].md")
    )
    for the_date, raw_content in sorted(normalized_files):
        single_stream.write(f"# {the_date.isoformat()}\n")
        single_stream.write(raw_content)
        single_stream.write("\n")

    return parse_memos_from_markdown(single_stream.getvalue())
