import datetime
import textwrap
from typing import Any, NotRequired, TypedDict

import marko
import marko.block
import yaml
from jinja2 import Environment


class Meta(TypedDict):
    author: str


class Header(TypedDict):
    headline: str
    motto: list[str]


class Article(TypedDict):
    date: datetime.date
    title: str
    content: str
    tags: NotRequired[list[str]]


class Memos(TypedDict):
    title: str
    meta: Meta
    header: Header
    articles: list[Article]


def parse_articles(renderer: marko.renderer.Renderer, document: marko.block.Document) -> list[Article]:
    """Parse a markdown document into a list of articles."""
    parser = _ArticleParser(renderer)
    return parser.parse(document)


class _ArticleParser:
    """Parser for extracting articles from markdown documents."""

    DATE_HEADER_LEVEL = 1
    ARTICLE_HEADER_LEVEL = 2

    def __init__(self, renderer: marko.renderer.Renderer) -> None:
        self.renderer = renderer
        self.articles: list[Article] = []
        self.current_date: datetime.date | None = None
        self.current_article: Article | None = None
        self.current_content_element = marko.block.Document()

    def parse(self, document: marko.block.Document) -> list[Article]:
        """Parse the document and return list of articles."""
        for child in document.children:
            if isinstance(child, marko.block.Heading):
                self._handle_heading(child)
            elif self.current_article is not None:
                self._add_content_block(child)

        self._finalize_current_article()
        return self.articles

    def _handle_heading(self, heading: marko.block.Heading) -> None:
        """Process a heading element."""
        self._finalize_current_article()

        if heading.level == self.DATE_HEADER_LEVEL:
            self._process_date_header(heading)
        elif heading.level == self.ARTICLE_HEADER_LEVEL:
            self._process_article_header(heading)

    def _process_date_header(self, heading: marko.block.Heading) -> None:
        """Process a date header (# YYYY-MM-DD)."""
        date_str = heading.children[0].children
        self.current_date = datetime.date.fromisoformat(date_str)

    def _process_article_header(self, heading: marko.block.Heading) -> None:
        """Process an article header (## Title #tag1 #tag2)."""
        if self.current_date is None:
            msg = "Article found without a date header"
            raise ValueError(msg)

        title_and_tags = heading.children[0].children
        title, tags = _parse_title_and_tags(title_and_tags)

        self.current_article = {
            "date": self.current_date,
            "title": title,
            "content": "",
        }
        if tags:
            self.current_article["tags"] = tags

    def _add_content_block(self, block: marko.element.Element) -> None:
        """Add a content block to the current article."""
        self.current_content_element.children.append(block)

    def _finalize_current_article(self) -> None:
        """Finalize and add the current article to the list."""
        if self.current_article is not None:
            self.current_article["content"] = self.renderer.render_children(self.current_content_element)
            self.articles.append(self.current_article)
            self.current_article = None
            self.current_content_element = marko.block.Document()


def _parse_title_and_tags(title_and_tags: str) -> tuple[str, list[str]]:
    """Parse 'Title #tag1 #tag2' into title and tags."""
    parts = marko.html_renderer.HTMLRenderer.escape_html(title_and_tags).split()
    tags = []
    title_parts = []

    for part in parts:
        if part.startswith("#"):
            tags.append(part[1:])  # Remove the # prefix
        else:
            title_parts.append(part)

    title = " ".join(title_parts)
    return title, tags


def parse_memos_from_markdown(source: str) -> Memos:
    expected_frontmatter_parts = 3

    whole_file = source.strip()
    parts = whole_file.split("---", maxsplit=2)
    if len(parts) < expected_frontmatter_parts:
        msg = "Missing frontmatter (the metadata between `---`)."
        raise ValueError(msg)
    _, raw_frontmatter, raw_content = parts

    engine = marko.Markdown()

    frontmatter = yaml.safe_load(raw_frontmatter)
    md_document = engine.parse(raw_content)
    content = parse_articles(engine.renderer, md_document)

    title = engine.parse(frontmatter["title"])
    title.children[0]._tight = True  # noqa: SLF001
    headline = engine.parse(frontmatter["header"]["headline"])
    headline.children[0]._tight = True  # noqa: SLF001
    mottos = [engine.parse(motto) for motto in frontmatter["header"]["motto"]]
    for motto in mottos:
        motto.children[0]._tight = True  # noqa: SLF001

    return {
        "title": engine.render(title),
        "meta": {"author": frontmatter["meta"]["author"]},
        "header": {"headline": engine.render(headline), "motto": [engine.render(motto) for motto in mottos]},
        "articles": content,
    }


def normalize_logseq_content(filename_stem: str, content: str) -> tuple[datetime.date, str]:
    the_date = datetime.date.fromisoformat(filename_stem.replace("_", "-"))

    # Get rid of tabs
    content = content.replace("\t", "    ")
    # Add some common indentation to titles so that they match with their content
    content_lines: list[str] = []
    for line in content.splitlines():
        match line:
            case _ if line.endswith("logseq.order-list-type:: number"):
                ordered_item = content_lines[-1]
                ordered_item = (
                    ordered_item.replace("-", "1.", 1)
                    if ordered_item.lstrip().startswith("-")
                    else f"    1. {ordered_item.lstrip()}"
                )
                content_lines[-1] = ordered_item
            case _ if line.startswith("##"):  # Title is the first line of the doc
                content_lines.append(f"    {line}")
            case _ if line.startswith("- ##"):  # A title later on
                content_lines.append(f"\n    {line[2:]}")
            case _ if line.startswith("    - "):  # A regular bullet point
                content_lines.append(f"\n    {line[6:]}")
            case _ if line.startswith("      "):  # A continued bullet point
                content_lines.append(f"    {line[6:]}")
            case _:
                content_lines.append(line)
    content = "\n".join(content_lines)
    # Remove common indentation
    content = textwrap.dedent(content)
    if content:
        # Add trailing newline
        content = content + "\n"

    return the_date, content


def month_day_ordinal_suffix(n: int) -> str:
    """Generate ordinal suffix for a day of month."""
    if 10 <= n <= 20:  # noqa: PLR2004
        return "th"
    return {1: "st", 2: "nd", 3: "rd"}.get(n % 10, "th")


def friendly_date(date: datetime.date) -> str:
    """Format date in friendly format (e.g., 'Friday, 29th October 2021')."""
    return date.strftime(f"%A, %-d{month_day_ordinal_suffix(date.day)} %B %Y")


def prepare_article_data(article: Article) -> dict[str, Any]:
    """Prepare article data for template rendering."""
    content = article["content"].strip()
    html_content = marko.convert(content).strip()

    # Prepare tags
    tags = article.get("tags", [])
    tag_links = [{"name": tag, "url": f"#{tag}"} for tag in tags]

    return {
        "date_iso": article["date"].isoformat(),
        "date_friendly": friendly_date(article["date"]),
        "title": article["title"],
        "content": html_content,
        "tags": tag_links,
        "has_tags": len(tags) > 0,
    }


def prepare_site_data(memos: Memos) -> dict[str, Any]:
    """Prepare all site data for template rendering."""
    # Prepare articles in reverse chronological order
    articles_data = [prepare_article_data(article) for article in reversed(memos["articles"])]

    return {
        "title": memos["title"],
        "author": memos["meta"]["author"],
        "headline": memos["header"]["headline"],
        "motto_lines": memos["header"]["motto"],
        "articles": articles_data,
    }


def render_site(memos: Memos, template_env: Environment) -> str:
    """Render complete site HTML from memos data."""
    # Prepare data
    site_data = prepare_site_data(memos)

    # Get main template
    template = template_env.get_template("main.html.j2")

    # Render
    return template.render(site_data)
